==> 1.7.5-dev.tar.gz: lib/src/widgets_live_keyed.dart <==
part of '../flutterx_live_data.dart';

typedef KeyedLiveData<K, V> = ObservableData<Map<K, ObservableData<V>>>;

class KeyedLiveDataBuilder<K, V> extends StatelessWidget {
  final ObservableData<K> dataKey;
  final KeyedLiveData<K, V> data;
  final ObservableData<V>? fallback;
  final LiveWidgetBuilder<V> builder;
  final Widget? placeholder;

  const KeyedLiveDataBuilder({
    super.key,
    required this.dataKey,
    required this.data,
    this.fallback = const ImmutableLiveData.late(),
    required this.builder,
    this.placeholder,
  });

  @override
  StatelessElement createElement() => _KeyedLiveDataBuilderElement<K, V>(this);

  @override
  Widget build(BuildContext context) {
    final data = (context as _KeyedLiveDataBuilderElement<K, V>)._data;
    return data != null && data.initialized ? builder(context, data._data) : (placeholder ?? const SizedBox.shrink());
  }

  @override
  void debugFillProperties(DiagnosticPropertiesBuilder properties) => super.debugFillProperties(properties
    ..add(DiagnosticsProperty<ObservableData<K>>('dataKey', dataKey))
    ..add(DiagnosticsProperty<KeyedLiveData<K, V>>('data', data))
    ..add(DiagnosticsProperty<ObservableData<V>>('fallback', fallback))
    ..add(ObjectFlagProperty<LiveWidgetBuilder<V>>.has('builder', builder)));
}

class _KeyedLiveDataBuilderElement<K, V> extends StatelessElement {
  ObservableData<V>? _data;
  Disposable? _observer;

  @override
  KeyedLiveDataBuilder<K, V> get widget => super.widget as KeyedLiveDataBuilder<K, V>;

  _KeyedLiveDataBuilderElement(KeyedLiveDataBuilder<K, V> super.widget);

  @override
  void mount(Element? parent, Object? newSlot) {
    _link(widget);
    super.mount(parent, newSlot);
  }

  @override
  void update(KeyedLiveDataBuilder<K, V> newWidget) {
    final oldWidget = widget;
    if (!identical(oldWidget.dataKey, newWidget.dataKey) || !identical(oldWidget.data, newWidget.data)) {
      _unlink(oldWidget);
      _link(newWidget);
    }
    super.update(newWidget);
  }

  @override
  void unmount() {
    _unlink(widget);
    super.unmount();
  }

  void _link(KeyedLiveDataBuilder<K, V> widget) =>
      _observer = observe2Forever<K, Map<K, ObservableData<V>>>(widget.dataKey, widget.data, _observe);

  void _unlink(KeyedLiveDataBuilder<K, V> widget) {
    _observer?.dispose();
    _observer = null;
    _data?.removeObserver(_onChanged);
    _data = null;
  }

  void _observe(K key, Map<K, ObservableData<V>> data) {
    final last = _data;
    final result = _data = data[key] ?? widget.fallback;
    if (last != null) {
      if (identical(last, result)) return;
      last.removeObserver(_onChanged);
    }
    if (result == null) throw StateError('data does not contains key $key and no fallback was provider');
    result.observeForever(_onChanged, dispatcher: Dispatcher.postFrame);
  }

  void _onChanged(_) => markNeedsBuild();
}

==> 1.9.6.tar.gz: lib/src/widgets_live_keyed.dart <==
part of '../flutterx_live_data.dart';

typedef KeyedLiveData<K, V> = ObservableData<Map<K, ObservableData<V>>>;

class KeyedLiveDataBuilder<K, V> extends StatelessWidget {
  final ObservableData<K> dataKey;
  final KeyedLiveData<K, V> data;
  final ObservableData<V>? fallback;
  final LiveWidgetBuilder<V> builder;
  final Widget? placeholder;

  const KeyedLiveDataBuilder({
    super.key,
    required this.dataKey,
    required this.data,
    this.fallback = const ImmutableLiveData.late(),
    required this.builder,
    this.placeholder,
  });

  @override
  StatelessElement createElement() => _KeyedLiveDataBuilderElement<K, V>(this);

  @override
  Widget build(BuildContext context) {
    final data = (context as _KeyedLiveDataBuilderElement<K, V>)._data;
    return data != null && data.initialized ? builder(context, data._data) : (placeholder ?? const SizedBox.shrink());
  }

  @override
  void debugFillProperties(DiagnosticPropertiesBuilder properties) => super.debugFillProperties(properties
    ..add(DiagnosticsProperty<ObservableData<K>>('dataKey', dataKey))
    ..add(DiagnosticsProperty<KeyedLiveData<K, V>>('data', data))
    ..add(DiagnosticsProperty<ObservableData<V>>('fallback', fallback))
    ..add(ObjectFlagProperty<LiveWidgetBuilder<V>>.has('builder', builder)));
}

class _KeyedLiveDataBuilderElement<K, V> extends StatelessElement {
  ObservableData<V>? _data;
  Disposable? _observer;

  @override
  KeyedLiveDataBuilder<K, V> get widget => super.widget as KeyedLiveDataBuilder<K, V>;

  _KeyedLiveDataBuilderElement(KeyedLiveDataBuilder<K, V> super.widget);

  @override
  void mount(Element? parent, Object? newSlot) {
    _link(widget);
    super.mount(parent, newSlot);
  }

  @override
  void update(KeyedLiveDataBuilder<K, V> newWidget) {
    final oldWidget = widget;
    if (!identical(oldWidget.dataKey, newWidget.dataKey) || !identical(oldWidget.data, newWidget.data)) {
      _unlink(oldWidget);
      _link(newWidget);
    }
    super.update(newWidget);
  }

  @override
  void unmount() {
    _unlink(widget);
    super.unmount();
  }

  void _link(KeyedLiveDataBuilder<K, V> widget) =>
      _observer = observe2Forever<K, Map<K, ObservableData<V>>>(widget.dataKey, widget.data, _observe);

  void _unlink(KeyedLiveDataBuilder<K, V> widget) {
    _observer?.dispose();
    _observer = null;
    _data?.removeObserver(_onChanged);
    _data = null;
  }

  void _observe(K key, Map<K, ObservableData<V>> data) {
    final last = _data;
    final result = _data = data[key] ?? widget.fallback;
    if (last != null) {
      if (identical(last, result)) return;
      last.removeObserver(_onChanged);
    }
    if (result == null) throw StateError('data does not contains key $key and no fallback was provider');
    result.observeForever(_onChanged, dispatcher: Dispatcher.endOfFrame);
  }

  void _onChanged(_) => markNeedsBuild();
}
